"use strict";
/**************************************************************************************************

	This software package is released under the Modified BSD License.

	Eclipsed Moon

		Copyright © 2017, Eurythrace Perseides. All rights reserved.

		An add-on/extension exclusively for the Pale Moon browser for smart, unobtrusive user
		agent spoofing.

										――――――――――――――――――――

										Modified BSD License

	Redistribution and use in source and binary forms, with or without modification, are permitted
	provided that the following conditions are met:

	1.	Redistributions of source code must retain the above copyright notices, this list of
		conditions and the following disclaimer.

	2.	Redistributions in binary form must reproduce the above copyright notice, this list of
		conditions and the following disclaimer in the documentation and/or other materials
		provided with the distribution.

	3.	Neither the name of the copyright holder nor the names of its contributors may be used to
		endorse or promote products derived from this software without specific prior written
		permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS	“AS IS” AND ANY EXPRESS OR
	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
	AND FITNESS FOR	A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
	SERVICES; LOSS OF USE, DATA, OR	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.

**************************************************************************************************/

const debugEclipsedMoon = false;

if( typeof( Cc ) === "undefined" ) { const Cc = Components.classes; }
if( typeof( Ci ) === "undefined" ) { const Ci = Components.interfaces; }
if( typeof( Cr ) === "undefined" ) { const Cr = Components.results; }
if( typeof( Cu ) === "undefined" ) { const Cu = Components.utils; }

//const { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

Cu.import( "resource://eclipsedmoon/modules/MersenneTwister.jsm" );
Cu.import( "resource://eclipsedmoon/modules/ForgetAboutSite.jsm" );

//Cu.import( "resource://gre/modules/AddonManager.jsm" );

//-------------------------------------------------------------------------------------------------
// Define the constructor for the user agent spoofer object. 

var UmbraCaster = ( function()
{
	var Observer				= Cc["@mozilla.org/observer-service;1"].getService( Ci.nsIObserverService ),
		Mediator				= Cc["@mozilla.org/appshell/window-mediator;1"].getService( Ci.nsIWindowMediator ),
		Timer					= Cc["@mozilla.org/timer;1"].createInstance( Ci.nsITimer ),
		IO_Service				= Cc["@mozilla.org/network/io-service;1"].getService( Ci.nsIIOService ),
		ScriptSecurityManager	= Cc["@mozilla.org/scriptsecuritymanager;1"].getService( Ci.nsIScriptSecurityManager ),
		PublicSuffixes			= Cc["@mozilla.org/network/effective-tld-service;1"].getService( Ci.nsIEffectiveTLDService ),
		PrefService				= Cc["@mozilla.org/preferences-service;1"].getService( Ci.nsIPrefService ),
		ePreferences			= PrefService.getBranch( "extensions.eclipsedmoon." ),
		gPreferences			= PrefService.getBranch( "general." ),
		MT						= new MersenneTwister(),
		SavedOverrides			= {};

	var uc = function() {};

//-------------------------------------------------------------------------------------------------
//	Initialize the spoofer object when the browser starts up.
	uc.prototype.startup = function()
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: startup: mode = “" + ePreferences.getCharPref( "mode" ) + "”." ); }
	//	Check if this the first run after installation.
		if( ePreferences.getCharPref( "mode" ) === "fresh" ) { this.install(); }

	//	Remove the listener that got us here - it is no longer needed.
		window.removeEventListener( "load", EclipsedMoon.startupHandler, false );

	//	Check operating status.
		if( ePreferences.getBoolPref( "state" ) )
		{
		//	Initialize everything.
			this.switchOn();
		}
		else
		{
		//	Shutdown everything.
			this.switchOff();
		}
	};

//-------------------------------------------------------------------------------------------------
//	Clean up the spoofer object before the browser exits.
	uc.prototype.shutdown = function()
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the shutdown routine." ); }
	//	Remove the listener that got us here - it is no longer needed.
		window.removeEventListener( "load", EclipsedMoon.shutdownHandler, false );

	//	Go dormant.
		this.switchOff();
	};

//-------------------------------------------------------------------------------------------------
//	Run this method only the first time after the addon has been installed.
//	This will automatically add the addon button to the Navigation Toolbar.
	uc.prototype.install = function()
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the install routine." ); }
		ePreferences.setCharPref( "mode", "smart" );
		
		const buttonID = "EclipsedMoon_toolbarbutton", toolbarID = "nav-bar";
		if( typeof( document.getElementById( buttonID ) ) === "undefined" )
		{
			let toolbar = document.getElementById( toolbarID );
			toolbar.insertItem( buttonID );
			toolbar.setAttribute( "currentset", toolbar.currentSet );
			document.persist( toolbarID, "currentset" );
		}
	};

//-------------------------------------------------------------------------------------------------
//	Run this method if the addon is being uninstalled.
	uc.prototype.uninstall = function() {};

//-------------------------------------------------------------------------------------------------
//	Define the common set of tasks necessary to activate the addon.
	uc.prototype.switchOn = function()
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the switchOn routine." ); }
		let nav = window.navigator;

	//	If none of the general override branches currently exist, define and initialize them with the native browser values.
		if( !gPreferences.prefHasUserValue( "buildID.override" ) ) { gPreferences.setCharPref( "buildID.override", nav.buildID ); }
		if( !gPreferences.prefHasUserValue( "oscpu.override" ) ) { gPreferences.setCharPref( "oscpu.override", nav.oscpu ); }
		if( !gPreferences.prefHasUserValue( "platform.override" ) ) { gPreferences.setCharPref( "platform.override", nav.platform ); }
		if( !gPreferences.prefHasUserValue( "useragent.override" ) ) { gPreferences.setCharPref( "useragent.override", nav.userAgent ); }

	//	Initialize the User Agent, OS/CPU, and Malformed Domain lists.
		this.initUAList();
		this.initOSList();
		this.initMDList();

	//	Add the onPageReadyStateChange handler.
		let appcontent = document.getElementById( "appcontent" );		// browser
		if( typeof( appcontent ) !== "undefined" )
		{
			this.readystatechangeHandler = this.onPageReadyStateChange.bind( this );
			appcontent.addEventListener( "readystatechange", this.readystatechangeHandler, true );
		}

	//	Add the preference change observers.
		gPreferences.addObserver( "useragent.override.", this, false );
		ePreferences.addObserver( "mode", this, false );
		ePreferences.addObserver( "device.laptop.oscpu", this, false );
		ePreferences.addObserver( "device.laptop.weights.os", this, false );
		ePreferences.addObserver( "device.laptop.useragent", this, false );
		ePreferences.addObserver( "device.laptop.weights.ua", this, false );

	//	Set the appropriate User Agent overrides.
		this.setOverrides();

	//	Show the current state.
		this.guiShowState( "on" );
	};

//-------------------------------------------------------------------------------------------------
//	Define the common set of tasks necessary to deactivate the addon.
	uc.prototype.switchOff = function()
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the switchOff routine." ); }
	//	Remove the onPageReadyStateChange handler.
		let appcontent = document.getElementById( "appcontent" );		// browser
		if( typeof( appcontent ) !== "undefined" )
		{
			appcontent.removeEventListener( "readystatechange", this.readystatechangeHandler, true );
		}

	//	Remove the preference change observers.
		try
		{
			gPreferences.removeObserver( "useragent.override.", this );
			ePreferences.removeObserver( "mode", this );
			ePreferences.removeObserver( "device.laptop.oscpu", this );
			ePreferences.removeObserver( "device.laptop.weights.os", this );
			ePreferences.removeObserver( "device.laptop.useragent", this );
			ePreferences.removeObserver( "device.laptop.weights.ua", this );
		}
		catch( er ){}

	//	Remove the User Agent overrides to allow the native values to be used.
		gPreferences.clearUserPref( "buildID.override" );
		gPreferences.clearUserPref( "oscpu.override" );
		gPreferences.clearUserPref( "platform.override" );
		gPreferences.clearUserPref( "useragent.override" );

	//	Show the current state.
		this.guiShowState( "off" );
	};

//-------------------------------------------------------------------------------------------------
//	Define the preference change observer.
	uc.prototype.observe = function( subject, topic, data )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Prefs Observer: “" + data + "” was changed." ); }
		if( topic !== "nsPref:changed" ) { return; }
		switch( data )
		{
		case "mode":
			this.setOverrides();
			return;
		case "device.laptop.oscpu":
		case "device.laptop.weights.os":
			this.initOSList();
		case "device.laptop.useragent":
		case "device.laptop.weights.ua":
			this.initUAList();
			return;
		default:
			this.initMDList();
		}
	};

//-------------------------------------------------------------------------------------------------
//	Save the current overrides.
	uc.prototype.saveOverrides = function()
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the saveOverrides routine." ); }
		SavedOverrides.buildID = "";
		if( gPreferences.prefHasUserValue( "buildID.override" ) ) { SavedOverrides.buildID = gPreferences.getCharPref( "buildID.override" ); }
		SavedOverrides.oscpu = "";
		if( gPreferences.prefHasUserValue( "oscpu.override" ) ) { SavedOverrides.oscpu = gPreferences.getCharPref( "oscpu.override" ); }
		SavedOverrides.platform = "";
		if( gPreferences.prefHasUserValue( "platform.override" ) ) { SavedOverrides.platform = gPreferences.getCharPref( "platform.override" ); }
		SavedOverrides.useragent = "";
		if( gPreferences.prefHasUserValue( "useragent.override" ) ) { SavedOverrides.platform = gPreferences.getCharPref( "useragent.override" ); }
	};

//-------------------------------------------------------------------------------------------------
//	Restore the previous overrides.
	uc.prototype.restoreOverrides = function()
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the restoreOverrides routine." ); }
		gPreferences.setCharPref( "buildID.override", SavedOverrides.buildID );
		gPreferences.setCharPref( "oscpu.override", SavedOverrides.oscpu );
		gPreferences.setCharPref( "platform.override", SavedOverrides.platform );
		gPreferences.setCharPref( "useragent.override", SavedOverrides.useragent );

		document.getElementById( "EclipsedMoon_toolbarbutton" ).setAttribute( "tooltiptext", SavedOverrides.useragent );
	};

//-------------------------------------------------------------------------------------------------
//	Set the general overrides.
	uc.prototype.setOverrides = function()
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the setOverrides routine." ); }
		let uaFull;
		if( ePreferences.getCharPref( "mode" ) === "static" )
		{
			uaFull = ePreferences.getCharPref( "staticUserAgent" );
		}
		else
		{
			uaFull = this.getRandomBOPU();
		}
		gPreferences.setCharPref( "buildID.override", this.getBuildID( uaFull ) );
		gPreferences.setCharPref( "oscpu.override", this.getOSCPU( uaFull ) );
		gPreferences.setCharPref( "platform.override", this.getPlatform( uaFull ) );
		let ua = this.getUserAgent( uaFull );
		gPreferences.setCharPref( "useragent.override", ua );

		document.getElementById( "EclipsedMoon_toolbarbutton" ).setAttribute( "tooltiptext", ua );
	};

//-------------------------------------------------------------------------------------------------
//	Define the initialization for the User Agent List.
	uc.prototype.initUAList = function( event )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the initUAList routine." ); }
	//	Initialize the list of User Agents.
		let list = ePreferences.getCharPref( "device.laptop.weights.os" ).split( "\n" ), wtOSXref = {}, wtUAXref = {};
		for( let i = 0, frags; i < list.length; i++ )
		{
			frags = list[i].split( /\s*=\s*/ );
			switch( frags[0] )
			{
			case "Windows":
				wtOSXref.WIN = { "regex": /WIN/, "weight": parseInt( frags[1], 10 ) };
				break;
			case "Mac":
				wtOSXref.MAC = { "regex": /MAC/, "weight":  parseInt( frags[1], 10 ) };
				break;
			case "Linux":
			default:
				wtOSXref.NUX = { "regex": /NUX/, "weight": parseInt( frags[1], 10 ) };
			}
		}

		list = ePreferences.getCharPref( "device.laptop.weights.ua" ).split( "\n" );
		for( let i = 0, frags; i < list.length; i++ )
		{
			frags = list[i].split( /\s*=\s*/ );
			wtUAXref[frags[0]] = { "regex": new RegExp( frags[0] ), "weight": parseInt( frags[1], 10 ) };
		}

		list = ePreferences.getCharPref( "device.laptop.useragent" ).split( "\n" );
		this.uaList = [];
		for( let i = 0, wtOS, wtUA, N; i < list.length; i++ )
		{
			wtOS = 1;
			wtUA = 1;
			if( list[i] === "" ) { continue; }			// Skip blank lines.
			if( /\/\//.test( list[i] ) ) { continue; }	// Skip comment lines.
			for( let os in wtOSXref )
			{
				if( wtOSXref[os].regex.test( list[i] ) )
				{
					wtOS = wtOSXref[os].weight;
					break;
				}
			}
			for( let ua in wtUAXref )
			{
				if( wtUAXref[ua].regex.test( list[i] ) )
				{
					wtUA = wtUAXref[ua].weight;
					break;
				}
			}
			N = wtOS * wtUA;
			if( N === 1 )
			{
				this.uaList.push( list[i] );
			}
			else
			{
				for( let j = 0; j < N; j++ )
				{
					this.uaList.push( list[i] );
				}
			}
		}

		// Fisher–Yates/Durstenfeld/Knuth Shuffle Algorithm

		for( let i = this.uaList.length - 1, prn = 0, tmp = ""; i > 0; i-- )
		{
			prn = Math.floor( MT.random() * ( i + 1 ) );
			tmp = this.uaList[i];
			this.uaList[i] = this.uaList[prn];
			this.uaList[prn] = tmp;
		}
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: initUAList: uaList Length: “" + this.uaList.length + "”." ); }
	};

//-------------------------------------------------------------------------------------------------
//	Define the initialization for the OS/CPU List.
	uc.prototype.initOSList = function( event )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the initOSList routine." ); }
		let list = ePreferences.getCharPref( "device.laptop.oscpu" ).split( "\n" );
		this.osList = { "WIN": [], "MAC": [], "NUX": [] };
		for( let i = 0; i < list.length; i++ )
		{
			if( /Windows/.test( list[i] ) ) { this.osList.WIN.push( list[i] ); continue; }
			if( /Mac/.test( list[i] ) ) { this.osList.MAC.push( list[i] ); continue; }
			if( /Linux/.test( list[i] ) ) { this.osList.NUX.push( list[i] ); continue; }
		}
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: initOSList: osList Length: “" + ( this.osList.WIN.length + this.osList.MAC.length + this.osList.NUX.length ) + "”." ); }
	};

//-------------------------------------------------------------------------------------------------
//	Define the initialization for the Malformed Domain List.
	uc.prototype.initMDList = function( event )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the initMDList routine." ); }
		this.mdList = { "length": 0 };
		var prefs = PrefService.getBranch( "general.useragent.override." );
		prefs.getChildList( "", {} ).forEach( function( domain, i, childArray )
		{
			this.mdList[domain] = prefs.getCharPref( domain ) + "|" + this.unkBuildID;
			this.mdList.length++;
		}, this );
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: initMDList: mdList Length: “" + this.mdList.length + "”." ); }
	};

//-------------------------------------------------------------------------------------------------
//	Swap the general overrides out if the current domain is in the Malformed Domain List.
	uc.prototype.mdCheckForSwap = function( domain )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the mdCheckForSwap routine." ); }
		let ua = this.mdList[domain];
		if( typeof( ua ) !== "undefined" )
		{
			this.saveOverrides();
			this.setOverrides( ua );
		}
	};

//-------------------------------------------------------------------------------------------------
//	Retrieve the domain if the provided nsIURI object refers to an external resource.
	uc.prototype.getWebDomain = function( oURI )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the getWebDomain routine." ); }
	//	See if this is an external/online page.
	//	oURI is a nsIURI object, not a string.
	//	Trying to access the nsIURI “host” atttribute throws
	//	a NS_ERROR_FAILURE for URIs like about:blank, etc.
	//	Otherwise, if the domain does not have a least
	//	ONE (1) “.”, it cannot be external.
		if( typeof( oURI ) !== "object" ){ return undefined; }
		try
		{
			let domain = oURI.host;
			if( domain.indexOf( "." ) === -1 ){ return undefined; }
			return domain;
		}
		catch( er )
		{
			return undefined;
		}
	};

//-------------------------------------------------------------------------------------------------
//	Determine how many open tabs are external URIs.
	uc.prototype.getExternalTabCount = function()
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the getExternalTabCount routine." ); }
		let osWindowEnumerator = Mediator.getEnumerator( "navigator:browser" ), nExternalTabs = 0;
		while( osWindowEnumerator.hasMoreElements() )
		{
			let osWindow = osWindowEnumerator.getNext();
			let tabs = osWindow.gBrowser.tabs;
			for( let i = 0, bro, domain; i < tabs.length; i++ )
			{
				bro = gBrowser.getBrowserForTab( tabs[i] );
				domain = this.getWebDomain( bro.currentURI );
				if( typeof( domain ) === "undefined" ){ continue; }
				nExternalTabs++;
			}
		}
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: getExternalTabCount = “" + nExternalTabs + "”." ); }
		return nExternalTabs;
	};

//-------------------------------------------------------------------------------------------------
//	Determine how many open tabs have the same URI.
	uc.prototype.getSameURICount = function( uri )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the getSameURICount routine." ); }
		let osWindowEnumerator = Mediator.getEnumerator( "navigator:browser" ), nSameURI = 0, wid = 0;
		while( osWindowEnumerator.hasMoreElements() )
		{
			let osWindow = osWindowEnumerator.getNext();
			let tabs = osWindow.gBrowser.tabs;
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: getSameURICount: Open Tab Count for OS Window[" + wid + "]: “" + tab.length + "”." ); }
			for( let i = 0, bro, domain; i < tabs.length; i++ )
			{
				bro = gBrowser.getBrowserForTab( tabs[i] );
				domain = this.getWebDomain( bro.currentURI );
				if( typeof( domain ) === "undefined" ){ continue; }
				if( bro.currentURI.spec === uri ) { nSameURI++; }
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: getSameURICount: Same URI: “" + tabURI.spec + "”, “" + uri + "”, “" + nSameURI + "”." ); }
			}
			wid++;
		}
		return nSameURI;
	};

//-------------------------------------------------------------------------------------------------
//	Determine how many open tabs have the same root domain as the input root domain.
	uc.prototype.getSameRootCount = function( root )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the getSameRootCount routine." ); }
		let osWindowEnumerator = Mediator.getEnumerator( "navigator:browser" ), nSameRoot = 0, wid = 0;
		while( osWindowEnumerator.hasMoreElements() )
		{
			let osWindow = osWindowEnumerator.getNext();
			let tabs = osWindow.gBrowser.tabs;
			for( let i = 0, bro, domain, tabRoot; i < tabs.length; i++ )
			{
				bro = gBrowser.getBrowserForTab( tabs[i] );
				domain = this.getWebDomain( bro.currentURI );
				if( typeof( domain ) === "undefined" ){ continue; }
				tabRoot = PublicSuffixes.getBaseDomain( bro.currentURI );
				if( tabRoot === root ) { nSameRoot++; }
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: getSameRootCount for OS Window[" + wid + "]: Same Domain: “" + tabRoot + "”, “" + root + "”, “" + nSameRoot + "”." ); }
			}
			wid++;
		}
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: getSameRootCount total: " + nSameRoot + "." ); }
		return nSameRoot;
	};

//-------------------------------------------------------------------------------------------------
//	Initialize a new page when its readyState starts changing during loading.
	uc.prototype.onPageReadyStateChange = function( event )
	{
		let doc = event.explicitOriginalTarget;	// Normally the page’s document object.
		while( doc.nodeName.toLowerCase() !== "#document" ) { doc = doc.parentNode; }
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the onPageReadyStateChange routine: state = “" + doc.readyState + "”, URI = “" + doc.documentURI + "”." ); }

	//	See if this is the top level document.
		try
		{
			if( doc.defaultView.self !== doc.defaultView.top ){ return; }
		}
		catch( er )
		{
			return;								// For cross-domain iframe exceptions.
		}

	//	See if this is an external/online page.
		let bro = gBrowser.getBrowserForDocument( doc );
		let domain = this.getWebDomain( bro.currentURI );
		if( typeof( domain ) === "undefined" ){ return; }

	//	See what state the page loading is at.
		switch( doc.readyState )
		{
	//	Do nothing - the new page is uninitialized. (Never seems to get fired)
		case "uninitialized":
			return;
	//	Do nothing - the new page is loading. (Never seems to get fired)
		case "loading":
			return;
		case "interactive":
	//	Same state as "DOMContentLoaded".
	//	See if the new page is in the malformed domain list.
	//	If so, save the current general buildID, oscpu, platform, and useragent overrides,
	//	and then set them to match the malformed domain user agent override.
	//	Add an "unload" event listener to the window of the new page.
			this.mdCheckForSwap( domain );
			doc.defaultView.onunload = this.onPageUnload.bind( this );
			return;
		case "complete":
	//	The new page is fully loaded, reset the general buildID, oscpu, and platform overrides
	//	back to match the current random user agent. Do it by setTimeout to allow the new page
	//	to do whatever snooping it is going to do first.
			if( typeof( this.mdList[domain] ) !== "undefined" )
			{
				window.setTimeout( this.restoreOverrides.bind( this ), 750 );
			}

			if( ePreferences.getCharPref( "mode" ) === "page" )
			{
				window.setTimeout( this.setOverrides.bind( this ), 1000 );
			}
		}
	};

//-------------------------------------------------------------------------------------------------
//	Clean up when unloading a page.
	uc.prototype.onPageUnload = function( event )
	{
		let doc = event.explicitOriginalTarget;	// Normally the page’s document object.
		while( doc.nodeName.toLowerCase() !== "#document" ) { doc = doc.parentNode; }
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the onPageUnload routine: URI = “" + doc.documentURI + "”." ); }

	//	See if this is the top level document.
		try
		{
			if( doc.defaultView.self !== doc.defaultView.top ){ return; }
		}
		catch( er )
		{
			return;								// For cross-domain iframe exceptions.
		}

	//	See if this unload is because of a new page load in the tab OR if the tab was closed.
		let tabURI, etN = 1, srN = 1;
		let bro = gBrowser.getBrowserForDocument( doc );
		if( bro === null || typeof( bro ) === "undefined" )
		{
		//	The tab was closed, so the tab and browser objects are unavailable.
			tabURI = IO_Service.newURI( doc.documentURI, doc.characterSet, null );
		}
		else
		{
		//	A new page is being loaded in the tab.
			tabURI = bro.currentURI;
			etN++;
			srN++;
		}

	//	See if this is an external/online page.
		let domain = this.getWebDomain( tabURI );
		if( typeof( domain ) === "undefined" ){ return; }

		let win = doc.defaultView;
		win.onunload = null;

		switch( ePreferences.getCharPref( "mode" ) )
		{
		case "fresh":
			ePreferences.setCharPref( "mode", "smart" );
		case "smart":
			let nExternalTabs = this.getExternalTabCount();
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: onPageUnload: External Tab Count: “" + nExternalTabs + "”." ); }
			if( nExternalTabs < etN )
			{
				this.setOverrides();
			}
			else
			{
				let homePageURI = PrefService.getBranch( "browser.startup." ).getCharPref( "homepage" ), nSameURI = 0;
				if( /^http/.test( homePageURI ) )
				{
					nSameURI = this.getSameURICount( homePageURI );
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: onPageUnload: Same URI Count = “" + nSameURI + "”, External Tab Count: “" + nExternalTabs + "”." ); }
					if( nExternalTabs === nSameURI )
					{
						this.setOverrides();
					}
				}
			}
		}

		let root = PublicSuffixes.getBaseDomain( tabURI );
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: onPageUnload: root domain = “" + root + "”." ); }
		let nSameRoot = this.getSameRootCount( root );
		if( ePreferences.getBoolPref( "smartErase" ) && nSameRoot < srN )
		{
			for( let i = 0, idoc, iURI; i < win.frames.length; i++ )
			{
				idoc = win.frames[i].document;
				iURI = IO_Service.newURI( idoc.documentURI, idoc.characterSet, null );
				ForgetAboutSite.removeDataFromDomain( PublicSuffixes.getBaseDomain( iURI ) );
			}
			ForgetAboutSite.removeDataFromDomain( root );
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: onPageUnload: Clearing all history for: “" + root + "”." ); }
		}
	};

//---------------------------------------------------------------------------------------------
//	Extract the BuildID from the raw new pseudorandom User Agent.
	uc.prototype.unkBuildID = "20170101987654";
	uc.prototype.getBuildID = function( ua )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the getBuildID routine." ); }
		let frags = ua.split( "|" );
		if( frags.length === 2 && /\d+/.test( frags[1] ) ) { return frags[1]; }
		return this.unkBuildID;
	};

//---------------------------------------------------------------------------------------------
//	Extract the OS/CPU from the raw new pseudorandom User Agent.
	uc.prototype.getOSCPU = function( ua )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the getOSCPU routine." ); }
		let regex = /;\srv:\d+\.\d+/;					// For browsers that include a revision number.

		let oscpu = ua.substring( ua.indexOf( "(" ) + 1, ua.indexOf( ")" ) ), frags;
		if( /Windows/.test( oscpu ) )
		{
			// Do nothing.
		}
		else if( /Macintosh/.test( oscpu ) )
		{
			oscpu = oscpu.replace( /Macintosh;\s/, "" );
		}
		else if( /Linux/.test( oscpu ) )
		{
			frags = oscpu.split( "; " );
			for( var i = 0; i < frags.length; i++ )		// Ignore X11 and/or revision number info.
			{
				oscpu = frags[i];
				if( /Linux/.test( oscpu ) ) { break; }
			}
		}
		else
		{
		//	Unknown OS/CPU, fake it with generic 64 bit Linux.
			oscpu = "Linux x86_64";
		}

		if( regex.test( oscpu ) )						// Remove any residual revision number info.
		{
			oscpu = oscpu.replace( regex, "" );
		}
		return oscpu;
	};

//---------------------------------------------------------------------------------------------
//	Extract the Platform from the raw new pseudorandom User Agent.
	uc.prototype.getPlatform = function( ua )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the getPlatform routine." ); }
		let platform = ua.substring( ua.indexOf( "(" ) + 1, ua.indexOf( ")" ) ), frags;
		if( /Windows/.test( platform ) )
		{
			if( /Win64/.test( platform ) )
			{
				platform = "Win64";
			}
			else
			{
				platform = "Win32";
			}
		}
		else if( /Macintosh/.test( platform ) )
		{
			platform = "MacIntel";
		}
		else if( /Linux/.test( platform ) )
		{
			frags = platform.split( "; " );
			for( var i = 0; i < frags.length; i++ )		// Ignore X11 and/or revision number info.
			{
				platform = frags[i];
				if( /Linux/.test( platform ) ) { break; }
			}
		}
		else
		{
		//	Unknown OS/CPU, fake it with generic 64 bit Linux.
			platform = "Linux x86_64";
		}

		return platform;
	};

//---------------------------------------------------------------------------------------------
//	Extract the User Agent part from the raw new pseudorandom User Agent.
	uc.prototype.getUserAgent = function( ua )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the getUserAgent routine." ); }
		let frags = ua.split( "|" );
		if( frags.length === 2 && /\d+/.test( frags[1] ) ) { return frags[0]; }
		return ua;
	};

//-------------------------------------------------------------------------------------------------
//	Get the next pseudorandom BuildID, OS/CPU, Platform, and User Agent.
	uc.prototype.getRandomBOPU = function()
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the getRandomBOPU routine." ); }
		let ua = this.uaList[Math.floor( MT.random() * this.uaList.length )], os;
		if( /WIN/.test( ua ) ) { os = "WIN"; }
		else if( /MAC/.test( ua ) ) { os = "MAC"; }
		else if( /NUX/.test( ua ) ) { os = "NUX"; }
		else
		{
			os = "WIN";
			ua = "Mozilla/5.0 (WIN; rv:51.0) Gecko/20100101 Firefox/51.0|20170125094131";
		}
		ua = ua.replace( os, this.osList[os][Math.floor( MT.random() * this.osList[os].length )] );
		return ua;
	};

//-------------------------------------------------------------------------------------------------
// GUI event handler for the Toolbar Button to toggle Eclipsed Moon on/off.

	uc.prototype.guiToggleState = function( event )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the guiToggleState routine." ); }
	//	Check state value
		if( ePreferences.getBoolPref( "state" ) )		// state: true = on, false = off
		{
			ePreferences.setBoolPref( "state", false );
			this.switchOff();
		}
		else
		{
			ePreferences.setBoolPref( "state", true );
			this.switchOn();
		}
	};

//-------------------------------------------------------------------------------------------------
// GUI helper function to display the state.

	uc.prototype.guiShowState = function( state )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the guiShowState routine." ); }
		let button = document.getElementById( "EclipsedMoon_toolbarbutton" ), regex = /EclipsedMoon-(on|off)/;
		button.className = button.className.replace( regex, "EclipsedMoon-" + state );

		let ua = window.navigator.userAgent;
		if( gPreferences.prefHasUserValue( "useragent.override" ) && gPreferences.getCharPref( "useragent.override" ) !== "" )
		{
			ua = gPreferences.getCharPref( "useragent.override" );
		}
		button.setAttribute( "tooltiptext", ua );
	};

//-------------------------------------------------------------------------------------------------
// GUI event handler to open the Eclipsed Moon preferences dialog window.

	uc.prototype.guiOpenPrefsWindow = function( event )
	{
if( debugEclipsedMoon ) { console.log( ">>>> EclipsedMoon: Entering the guiOpenPrefsWindow routine." ); }
		if( typeof( this.prefsWindow ) === "undefined" || this.prefsWindow.closed )
		{
			let instantApply = PrefService.getBranch( "browser.preferences." ).getBoolPref( "instantApply" );
			let features = "centerscreen, chrome, resizable, titlebar, toolbar" + ( ( instantApply.value ) ? ", dialog=no" : ", modal" );
			this.prefsWindow = window.openDialog( "chrome://eclipsedmoon/content/EclipsePrefs.xul", "Eclipsed Moon Preferences", features );
		}
		this.prefsWindow.focus();
	};

//-------------------------------------------------------------------------------------------------
//	All done with construction.
	return uc;
} )();

var EclipsedMoon = new UmbraCaster();
EclipsedMoon.startupHandler = EclipsedMoon.startup.bind( EclipsedMoon );
EclipsedMoon.shutdownHandler = EclipsedMoon.shutdown.bind( EclipsedMoon );

window.addEventListener( "load", EclipsedMoon.startupHandler, false );
window.addEventListener( "unload", EclipsedMoon.shutdownHandler, false );
